#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" This is the entrypoint for ingestion launching """
import json
import logging
import sys
import tarfile
from pathlib import Path

from pycapo import CapoConfig

# pylint: disable=C0301, E0401, R0903, R1721
from typing import List, Tuple

import arrow
from ingest_envoy.manifest_components import (
    CURATOR_MANIFEST_FILENAME,
    INGESTION_ARTIFACTS_NAME,
    INIT_WEBLOG_FILENAME,
    JSON,
    MANIFEST_FILENAME,
    TARFILE_EXT,
    WEBLOG_FILENAME,
    AncillaryProduct,
    IngestionManifestKey,
    InputGroup,
    InputScienceProduct,
    ManifestComponentIF,
    ManifestParameters,
    OutputGroup,
    OutputScienceProduct,
    ReingestGroup,
)
from ingest_envoy.schema import AbstractTextFile
from ingest_envoy.std_img_manifest_utils import ImageIngestionProductsFinder
from ingest_envoy.std_obs_manifest_utils import ObservationIngestionProductsFinder
from ingest_envoy.utilities import (
    AncillaryProductType,
    CuratorType,
    IngestionManifestException,
    NoScienceProductException,
    ScienceProductType,
    Telescope,
    find_output_tars,
    find_weblogs,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))

# pylint: disable=R0902, R0913


class IngestionManifest(ManifestComponentIF):
    """Encapsulates an ingestion manifest"""

    def __init__(
        self,
        telescope: Telescope,
        sp_type: ScienceProductType,
        staging_source_dir: Path | None,  # curation has no staging directory
        locator: str | None,  # initial observation ingestion has no locator
        # if running curator, use ReingestGroup, otherwise none
        reingest_group: ReingestGroup | None,
        # all except Partial Curation, EVLA_EB and VLASS catalog manifest have input group
        input_group: InputGroup | None,
        # every non-curator manifest has at least one output group
        output_group: OutputGroup | None,
        # image manifest has this:
        additional_metadata: AbstractTextFile = None,
        filename: str = None,
    ):
        self.staging_source_dir = staging_source_dir
        self.sp_type = sp_type
        self.locator = locator
        self.reingest = reingest_group
        self.input_group = input_group
        self.output_group = output_group
        self.telescope = telescope
        self.filename = filename

        # Check if NGAS ingestion should be enabled for all manifests in this environment
        self.ngas_ingest = self.get_ngas_flag()

        self.parameters = self.build_ingest_parameters(additional_metadata)
        if staging_source_dir is not None:
            # we are not running curation, find files
            self.files_found = [file for file in self.staging_source_dir.iterdir()]

    def get_ngas_flag(self) -> bool:
        """Determine and retrieve the correct ingestNGAS flag for this type of ingestion/curation"""

        match self.sp_type:
            case ScienceProductType.EXEC_BLOCK:
                flag = CapoConfig().getboolean("archive-ingestion.ingestNGAS.observation")
            case ScienceProductType.EVLA_CAL:
                flag = CapoConfig().getboolean("archive-ingestion.ingestNGAS.calibration")
            case ScienceProductType.IMAGE | ScienceProductType.VLASS_SECI:
                flag = CapoConfig().getboolean("archive-ingestion.ingestNGAS.imaging")
            case _:
                flag = True

        return flag

    def build_ingest_parameters(self, additional_metadata: AbstractTextFile):
        """
        Make the "parameters" section of the manifest

        :return:
        """
        if self.sp_type not in [
            ScienceProductType.EVLA_CAL,
            ScienceProductType.IMAGE,
            ScienceProductType.VLASS_SECI,
            ScienceProductType.EXEC_BLOCK,
        ]:
            raise NotImplementedError()

        if self.reingest is not None:
            params = ManifestParameters(
                telescope=self.telescope,
                ngas_ingest=False,
                staging_source_dir=self.staging_source_dir,
            )

        elif additional_metadata:
            params = ManifestParameters(
                telescope=self.telescope,
                ngas_ingest=self.ngas_ingest,
                calibrate=False,
                staging_source_dir=self.staging_source_dir,
                additional_metadata=additional_metadata,
            )

        elif self.sp_type == ScienceProductType.IMAGE:
            params = ManifestParameters(
                telescope=self.telescope,
                ngas_ingest=self.ngas_ingest,
                calibrate=False,
                staging_source_dir=self.staging_source_dir,
            )

        else:
            params = ManifestParameters(
                telescope=self.telescope,
                ngas_ingest=self.ngas_ingest,
                staging_source_dir=self.staging_source_dir,
            )

        return params

    def __eq__(self, other):
        if isinstance(other, IngestionManifest):
            return (
                other.parameters == self.parameters
                and other.input_group == self.input_group
                and other.output_group == self.output_group
            )

        return False

    def ingestion_path(self) -> Path:
        """
        Get the ingestion path associated with this manifest.

        :return:
        """
        return self.parameters.staging_source_dir

    def write(self) -> Path:
        """
        Write the manifest .json file.

        :return:
        """
        me_dict = self.to_dict()
        if self.reingest is not None and self.reingest.targets is not None:
            output_path = Path.cwd() / CURATOR_MANIFEST_FILENAME
        elif self.reingest is not None:
            output_path = self.staging_source_dir / CURATOR_MANIFEST_FILENAME
        else:
            output_path = self.staging_source_dir / MANIFEST_FILENAME

        to_write = json.dumps(me_dict, indent=4)
        with open(output_path, "w") as out:
            out.write(to_write)

        return output_path

    def to_dict(self) -> JSON:
        """
        Turn this maifest into a Dictionary suitable for converting to JSON

        :return:
        """
        me_dict = dict(self.__dict__)

        to_return = {
            IngestionManifestKey.PARAMETERS.value: self.parameters.to_dict(),
        }

        # are we running curator?
        if IngestionManifestKey.REINGEST.value in me_dict and me_dict[IngestionManifestKey.REINGEST.value] is not None:
            to_return[IngestionManifestKey.REINGEST.value] = me_dict[IngestionManifestKey.REINGEST.value].to_dict()

        # partial-curation manifests have no output groups
        if (
            IngestionManifestKey.OUTPUT_GROUP.value in me_dict
            and me_dict[IngestionManifestKey.OUTPUT_GROUP.value] is not None
        ):
            to_return[IngestionManifestKey.OUTPUT_GROUP.value] = me_dict[
                IngestionManifestKey.OUTPUT_GROUP.value
            ].to_dict()

        if (
            IngestionManifestKey.INPUT_GROUP.value in me_dict
            and me_dict[IngestionManifestKey.INPUT_GROUP.value] is not None
        ):
            # Ingestion manifests with empty input groups can cause errors on ingest
            if len(self.input_group.science_products) > 0:
                to_return[IngestionManifestKey.INPUT_GROUP.value] = me_dict[
                    IngestionManifestKey.INPUT_GROUP.value
                ].to_dict()

        return to_return


class IngestionManifestBuilder:
    """Builds ingestion manifest and associated files from files in ingestion_path"""

    def __init__(
        self,
        staging_source_dir: Path | None,  # partial curation has no source path
        sp_type: str,
        locator: str | None,  # initial observation ingestion has no locator
        telescope: str,
        additional_metadata: AbstractTextFile = None,
        filename: str = None,
        curate: (CuratorType, str, List[str]) = None,
    ):
        # get the telescope
        self.telescope = Telescope(telescope)
        if self.telescope is None:
            raise ValueError("telescope is required.")

        # get the directory containing the files to be ingested
        self.staging_source_dir = staging_source_dir

        # get the additional metadata, if any (EVLA CAL ingestion will have none)
        self.additional_metadata = additional_metadata

        # are we running curator?
        self.curation_type = curate[0] if curate is not None else None

        # determine the science product type
        self.sp_type = ScienceProductType(sp_type)
        if self.sp_type is None or not isinstance(self.sp_type, ScienceProductType):
            raise ValueError(f"Invalid/unknown science product type: {sp_type}")
        if self.sp_type not in [
            ScienceProductType.EVLA_CAL,
            ScienceProductType.IMAGE,
            ScienceProductType.VLASS_SECI,
            ScienceProductType.EXEC_BLOCK,
        ]:
            raise NotImplementedError(f"Don't know yet how to build a {self.sp_type.value} manifest")

        if self.curation_type is not None:
            if self.curation_type not in [CuratorType.PARTIAL, CuratorType.FULL]:
                raise NotImplementedError(
                    f"Don't know how to build a {self.curation_type.value} curation {self.sp_type} manifest"
                )
            self.curation_source = Path(curate[1]) if curate and curate[1] else None
            self.target_list = curate[2]
            if self.curation_type == CuratorType.FULL and self.target_list != None:
                raise IngestionManifestException(
                    f"Expected targets=None for full curation, got targets={self.target_list}"
                )

        if locator is not None:
            # we are not running observation ingestion, use a locator
            self.locator = locator

        if staging_source_dir is not None:
            # we are not running curation, look for files
            self.files_found = [file for file in staging_source_dir.iterdir()]
            if len(self.files_found) == 0:
                raise IngestionManifestException(f"No ingestion files found at {staging_source_dir}")

        if filename is not None:
            self.filename = filename

    def build(self) -> Tuple[IngestionManifest, Path]:
        """
        Using only -relevant- files in ingestion_path, write the manifest
        and produce other files required for ingestion.

        :return: the ingestion manifest and the file containing its JSON
        """
        if self.curation_type == CuratorType.PARTIAL:
            return self._build_partial_curation_manifest()

        if self.sp_type == ScienceProductType.EVLA_CAL:
            return self._build_evla_cal_manifest()
        elif self.sp_type == ScienceProductType.EXEC_BLOCK:
            return self._build_observation_manifest(self.curation_type == CuratorType.FULL)

        return self._build_image_manifest()

    def _build_partial_curation_manifest(self) -> (IngestionManifest, Path):
        """
        Build a manifest for partial curator
        Partial curation is simple, only parameters and reingest groups are needed

        :return: the manifest file object and path
        """
        manifest = IngestionManifest(
            telescope=self.telescope,
            staging_source_dir=self.curation_source,
            locator=self.locator,
            sp_type=self.sp_type,
            reingest_group=self._build_reingest_group(),
            input_group=None,
            output_group=None,
        )

        manifest_file = manifest.write()

        return manifest, manifest_file

    def _build_evla_cal_manifest(self) -> (IngestionManifest, Path):
        # create the manifest
        manifest = IngestionManifest(
            telescope=self.telescope,
            locator=self.locator,
            sp_type=self.sp_type,
            staging_source_dir=self.staging_source_dir,
            reingest_group=None,
            input_group=self._build_input_group(),
            output_group=self._build_evla_cal_output_group(),
        )

        # We can't create the ingestion artifacts tar quite yet,
        # because it will contain the yet-to-be-written manifest itself
        # (required for ingestion, evidently)
        artifacts_ap = self._build_artifacts_product()
        if artifacts_ap not in manifest.output_group.ancillary_products:
            manifest.output_group.ancillary_products.append(artifacts_ap)

        if not manifest.output_group.ancillary_products:
            manifest.output_group.ancillary_products = []
        weblog_ap = AncillaryProduct(type=AncillaryProductType.PIPELINE_WEBLOG, filename=WEBLOG_FILENAME)
        if weblog_ap not in manifest.output_group.ancillary_products:
            manifest.output_group.ancillary_products.append(weblog_ap)

        manifest_file = manifest.write()
        self.write_ingestion_artifacts_tar(self.staging_source_dir / artifacts_ap.filename)

        return manifest, manifest_file

    def _build_observation_manifest(self, is_full_curation: bool = False) -> tuple[IngestionManifest, Path]:
        reingest_group = None
        if is_full_curation:
            reingest_group = self._build_reingest_group()
        # create the manifest
        manifest = IngestionManifest(
            telescope=self.telescope,
            locator=None,
            sp_type=self.sp_type,
            staging_source_dir=self.source_dir,
            reingest_group=reingest_group,
            input_group=InputGroup([]),
            output_group=self._build_observation_output_group(),
            filename=self.filename,
        )

        if not manifest.output_group.ancillary_products:
            manifest.output_group.ancillary_products = []
        artifacts_ap = self._build_artifacts_product()
        if artifacts_ap not in manifest.output_group.ancillary_products:
            manifest.output_group.ancillary_products.append(artifacts_ap)

        manifest_file = manifest.write()
        self.write_ingestion_artifacts_tar(self.source_dir / artifacts_ap.filename)

        return manifest, manifest_file

    def _find_init_weblog_if_any(self):
        """
        Is there an initial weblog in the staging source dir?
        (If so, this calibration is v2 or higher)

        :return: initial weblog, if any
        """
        for file in [file for file in self.staging_source_dir.iterdir()]:
            if file.name == INIT_WEBLOG_FILENAME:
                return file

        return None

    def _build_image_manifest(self) -> (IngestionManifest, Path):
        """
        Image manifest has additional_metadata, and output group is way more complicated
        :return:
        """

        # create the manifest
        manifest = IngestionManifest(
            telescope=self.telescope,
            locator=self.locator,
            additional_metadata=self.additional_metadata,
            sp_type=self.sp_type,
            staging_source_dir=self.staging_source_dir,
            reingest_group=None,
            input_group=self._build_input_group(),
            output_group=self._build_imaging_output_group(),
        )

        artifacts_ap = self._build_artifacts_product()
        if artifacts_ap not in manifest.output_group.ancillary_products:
            manifest.output_group.ancillary_products.append(artifacts_ap)
        manifest_file = manifest.write()
        self.write_ingestion_artifacts_tar(self.staging_source_dir / artifacts_ap.filename)

        return manifest, manifest_file

    def _build_reingest_group(self) -> ReingestGroup:
        """
        Create the curation reingest group

        :return:
        """
        return ReingestGroup(
            locator=self.locator,
            product_type=self.sp_type,
            targets=self.target_list,
        )

    def _build_input_group(self) -> InputGroup:
        """
        Create the input group using the parameters.

        :return:
        """

        # N.B. this is sufficient for most types of ingestion,
        # but ALMA CALs will have multiple EB SPs, identified only by locator,
        # and VLBAs have no input group at all.
        sp_in = InputScienceProduct(locator=self.locator)

        return InputGroup([sp_in])

    def _build_evla_cal_output_group(self) -> OutputGroup:
        """
        Create EVLA standard calibration manifest output group using the parameters
        and the contents of the staging dir.

        :return: an output group, if a science product is found
        """

        # find science product (we expect just one for this SP type)
        tars_found = find_output_tars(self.files_found, self.staging_source_dir)

        sci_prod = None
        for file in tars_found:
            sci_prod = OutputScienceProduct(product_type=self.sp_type, filename=file.name)
            break

        if sci_prod:
            weblog_files = find_weblogs(self.files_found, self.staging_source_dir)
            weblogs = []
            for file in weblog_files:
                ap = AncillaryProduct(type=AncillaryProductType.PIPELINE_WEBLOG, filename=file.name)
                if ap not in weblogs:
                    weblogs.append(ap)
        else:
            raise NoScienceProductException(f">>> NO SCIENCE PRODUCT FOUND in {self.staging_source_dir}")

        return OutputGroup(science_products=[sci_prod], ancillary_products=weblogs)

    def _build_imaging_output_group(self) -> OutputGroup:
        """
        Create imaging manifest output group using the parameters
        and the contents of the staging dir.

        :return:
        """

        products_finder = ImageIngestionProductsFinder(self.staging_source_dir, self.sp_type)
        science_products = products_finder.output_science_products
        ancillary_products = products_finder.ancillary_products

        return OutputGroup(science_products=science_products, ancillary_products=ancillary_products)

    def _build_observation_output_group(self) -> OutputGroup:
        """
        Create observation manifest output group using the parameters
        and the contents of the staging dir.

        :return:
        """

        products_finder = ObservationIngestionProductsFinder(self.source_dir, self.sp_type)
        science_products = products_finder.output_science_products
        ancillary_products = products_finder.ancillary_products

        return OutputGroup(science_products=science_products, ancillary_products=ancillary_products)

    @staticmethod
    def _build_artifacts_filename() -> str:
        """
        Build unique manifest filename in standard format.

        :return: the filename
        """
        current_time = arrow.now()
        timestamp = format_timestamp(current_time)
        return f"{INGESTION_ARTIFACTS_NAME}{timestamp}{TARFILE_EXT}"

    @staticmethod
    def _build_artifacts_product() -> AncillaryProduct:
        return AncillaryProduct(
            AncillaryProductType.INGESTION_ARTIFACTS, IngestionManifestBuilder._build_artifacts_filename()
        )

    def write_ingestion_artifacts_tar(self, artifacts_path: Path) -> tarfile.TarFile:
        """
        Take the list of files and write a tar file for inclusion in the archive.
        This happens in the staging area for ingestion.
        The EVLA CAL tar will contain just the manifest.

        :param artifacts_path: Path to create the resulting tar file at
        :return: tar file of the ingestion artifacts
        """

        addl_md_file = None
        if self.additional_metadata:
            # find the additional metadata
            addl_md_filename = self.additional_metadata.filename
            addl_md_file = self.source_dir / addl_md_filename

        with tarfile.open(artifacts_path, "w") as ingestion_artifacts_tar:
            if addl_md_file:
                ingestion_artifacts_tar.add(addl_md_file)

            # The manifest file itself is considered an ingestion artifact.
            # (It's turtles all the way down.)
            manifest_file = self.source_dir / MANIFEST_FILENAME
            if not manifest_file.exists():
                manifest_file = self.source_dir / CURATOR_MANIFEST_FILENAME
            if not manifest_file.exists():
                raise FileNotFoundError(
                    f"No manifest (i.e. {MANIFEST_FILENAME} or {CURATOR_MANIFEST_FILENAME}) found in {self.source_dir}"
                )
            ingestion_artifacts_tar.add(manifest_file)

        return ingestion_artifacts_tar

    @property
    def source_dir(self) -> Path:
        source_dir = self.staging_source_dir if self.staging_source_dir is not None else self.curation_source
        if source_dir is None:
            raise IngestionManifestException(
                "IngestionManifestBuilder needs a value for either staging_source_dir or curation_source, found neither"
            )
        return source_dir


def format_timestamp(datetime: arrow.Arrow) -> str:
    """
    Format the current time as
        2021_07_01T13_49_17.237

    :param datetime: current timestamp
    :return: timestamp suitable for ingestion manifest filename
    """

    return datetime.format("YYYY_MM_DDThh_mm_ss.SSS")


def find_manifest(ingestion_path: Path) -> Path:
    """
    Find the ingestion manifest at this ingestion path.

    :param ingestion_path: home of ingestion files
    :return:
    """
    for json_file in ingestion_path.glob(MANIFEST_FILENAME):
        return json_file

    raise FileNotFoundError(f"No ingestion manifest found at {ingestion_path}")
